/******************************************************************************
 * Circular buffer for 8-bit microcontrollers (and possibly others)
 *
 * Written in 2007 by Matt Mets (matt.mets@cibomahto.com)
 * This code is public domain- do with it what you would like.  If you find it
 * useful, please let me know!
 *****************************************************************************/


#ifndef CIRCULAR_BUFFER
#define CIRCULAR_BUFFER

#include <string.h>     /* for memcpy() */
#include "circular_buffer.h"


cbuffer_return_t cbuffer_init(cbuffer_buff_t *buff) {
    unsigned char i;

    buff->read_pos = 0;
    buff->write_pos = 0;

    return CBUFFER_MSG_SUCCESS;
}

unsigned char cbuffer_hasData(cbuffer_buff_t *buff) {
    if(buff->read_pos == buff->write_pos)
        return CBUFFER_MSG_NO_DATA;
    else 
        return CBUFFER_MSG_HAS_DATA;
}

unsigned char cbuffer_get_freeSpace(cbuffer_buff_t *buff) {
    if(buff->read_pos == buff->write_pos)
        return BUFFER_SIZE - 1;
    if((buff->read_pos == 0 && buff->write_pos == BUFFER_SIZE - 1) ||
       (buff->read_pos - buff->write_pos == 1))
        return 0;
    
    if(buff->read_pos < buff->write_pos)
        return BUFFER_SIZE - buff->write_pos + buff->read_pos - 1;
    else /* (read_pos > write_pos) */
        return buff->read_pos - buff->write_pos - 1;
}

cbuffer_return_t cbuffer_read_c(cbuffer_buff_t *buff, char *data) {
    if(buff->read_pos == buff->write_pos)
        return CBUFFER_MSG_NO_DATA;
	    
    *data = buff->buffer[ buff->read_pos ];
    buff->read_pos = (buff->read_pos + 1) & (BUFFER_SIZE -1);
    return CBUFFER_MSG_SUCCESS;
}

cbuffer_return_t cbuffer_write_c(cbuffer_buff_t *buff, char data) {
    if(cbuffer_get_freeSpace(buff) == 0)
        return CBUFFER_MSG_NOT_ENOUGH_SPACE;
    
    buff->buffer[buff->write_pos] = data;
    buff->write_pos = (buff->write_pos + 1) & (BUFFER_SIZE - 1);
    return CBUFFER_MSG_SUCCESS;
}

/* TODO: Finish writing this function. */
#if 0
cbuffer_return_t read_string(cbuffer_buff_t *buff,
                             char *data, unsigned char *length) {
    /* length is the maximum length to read bach (size of *data) on call
     * and on successful read is overwritten with the acutal length of data
     * read.
     */
    if(buff->read_pos == buff->write_pos)
        return CBUFFER_MSG_NO_DATA;
   
    /* check to see if the data wraps; if it doesnt, copy in one fell sweep,
     * otherwise copy in two calls
     */
 
    return CBUFFER_MSG_ERROR;
}
#endif

cbuffer_return_t cbuffer_write_s(cbuffer_buff_t *buff, char *data, unsigned char length) {
    size_t length_a, length_b;
    char a, b;

    if(length > (BUFFER_SIZE - 1))
        return CBUFFER_MSG_WOULD_NEVER_FIT;

    if(length > cbuffer_get_freeSpace(buff))
        return CBUFFER_MSG_NOT_ENOUGH_SPACE;

    length_a = length;  /* This is because memcpy requires a size_t (which is
                         * declared as unsiged short long int by the C18
                         * compiler), but we use a char for the length to
                         * make things faster (the buffer can be at max 256
                         * characters long).
                         */

    if(buff->write_pos < buff->read_pos) {
        /* In this case, the free space does not wrap across the boundary and
         * the data can be copied intact.
         */
        memcpy(&buff->buffer[buff->write_pos], (void*) data, length_a);
        buff->write_pos = (buff->write_pos + length) & (BUFFER_SIZE - 1);
    }
    else {
        if(length < (BUFFER_SIZE - buff->write_pos)) {
            /* If the data does not wrap across the boundary, copy it in one
             * motion.
             * note: we don't need to check for the case where we are writing
             * data to the end of the buffer and read_pos = 0, because that
             * would mean the data being copied is too big for the buffer.
             */
            memcpy(&buff->buffer[buff->write_pos], (void*) data, length_a);
            buff->write_pos = (buff->write_pos + length) & (BUFFER_SIZE - 1);
        }
        else {
            /* fill to the end of the buffer, then copy the remainder to the
             * beginning of the buffer.
             */
            length_a =  (BUFFER_SIZE - buff->write_pos);
            length_b = (length - BUFFER_SIZE + buff->write_pos);

            memcpy(&buff->buffer[buff->write_pos], (void*) data, length_a);
            memcpy(&buff->buffer[0], (void*) &data[length_a], length_b); 
            buff->write_pos = length_b;
        }
    }

    return CBUFFER_MSG_SUCCESS;
}

#endif /* CIRCULAR_BUFFER */
