/******************************************************************************
 * Circular buffer for 8-bit microcontrollers (and possibly others)
 *
 * Written in 2007 by Matt Mets (matt.mets@cibomahto.com)
 * This code is public domain- do with it what you would like.  If you find it
 * useful, please let me know!
 *****************************************************************************/

/** This defines the size of all buffers.  It must be a multiple of two and the
 * greatest value it be is 256.
 */
#define BUFFER_SIZE 256

/** Return status of the functions */
typedef enum cbuffer_return {
    CBUFFER_MSG_SUCCESS,            /* Function completed successfully */ 
    CBUFFER_MSG_NO_DATA,            /* Buffer is empty */
    CBUFFER_MSG_HAS_DATA,           /* Buffer has data in it */
    CBUFFER_MSG_NOT_ENOUGH_SPACE,   /* The buffer is too full right now to
                                     * add this data.
                                     */
    CBUFFER_MSG_WOULD_NEVER_FIT,    /* This data is larger then the maximum
                                     * size of the buffer and would never fit
                                     */
    CBUFFER_MSG_ERROR               /* Unspecified error */
} cbuffer_return_t;

/** The structure which will hold the buffer.  This is not to be used directly
 * by the user program, but rather indirectly through the accessor functions.
 */
typedef struct cbuffer_buff {
    unsigned char read_pos,         /* Index where reading should take place */
                  write_pos;        /* Index where writing should take place */
    char buffer[BUFFER_SIZE];       /* Array to store the data in */
} cbuffer_buff_t;

/** @name cbuffer_init
 * Init the buffer, must be called before using the buffer
 * @param buff  The buffer to initialize
 * @return returns CBUFFER_MSG_SUCCESS
 */
cbuffer_return_t cbuffer_init(cbuffer_buff_t *buff);

/** @name cbuffer_hasData
 * Check to see if the buffer has data in it
 * @param buff The buffer to check
 * @return Returns CBUFFER_MSG_HAS_DATA or CBUFFER_MSG_NO_DATA
 * */
unsigned char cbuffer_hasData(cbuffer_buff_t *buff);

/** @name cbuffer_get_freeSpace
 * Get the free space left in the buffer
 * @param buff The buffer to check
 * @return Returns the free space left in the buffer, in bytes
 */
unsigned char cbuffer_get_freeSpace(cbuffer_buff_t *buff);

/** @name cbuffer_write_c
 * Write a single character to the buffer
 * @param buff The buffer to write to
 * @param data Single character to add to the buffer
 * @return Returns CBUFFER_MSG_SUCCESS on a successful write
 */
cbuffer_return_t cbuffer_write_c(cbuffer_buff_t *buff, char data);

/** @name cbuffer_write_s
 * Write a string to the buffer
 * @param buff The buffer to write to
 * @param data Data string to add to the buffer
 * @param len Length of the string to be added
 * @return Returns CBUFFER_MSG_SUCCESS on a successful write
 */
cbuffer_return_t cbuffer_write_s(cbuffer_buff_t *buff,
                                 char *data, unsigned char len);

/** @name cbuffer_read_c
 * Read a single character from the buffer and remove it from the buffer
 * @param buff The buffer to read from
 * @param data The location to write the data to
 * @return Returns CBUFFER_MSG_SUCCESS on a successful read
 */
cbuffer_return_t cbuffer_read_c(cbuffer_buff_t *buff, char *data);

