/******************************************************************************
 * An RS232-based PWM controller for the PIC18F2680 microcontroller
 * Compile with the MPLAB C18 compiler
 *
 * Written in 2007 by Matt Mets (matt.mets@cibomahto.com)
 * This code is public domain- do with it what you would like.
 *****************************************************************************/


/***** 18F2680 Configuration *************************************************/
#include <p18f2680.h>
#pragma config OSC = IRCIO67	/* Use internal oscillator */
#pragma config WDT = OFF	    /* Turn off watchdog timer */
#pragma config LVP = OFF	    /* Disable low voltage programming */
#pragma config DEBUG = OFF	    /* Disable debug mode */

/***** Other Includes ********************************************************/
#include <usart.h>
#include "circular_buffer.h"

/***** Macros and defines ****************************************************/
/* shortcuts for each of the I/O pins */
#define LED         PORTAbits.RA3

#define PWM_NUM_CHANNELS 8

#define pwm_channel_0  PORTBbits.RB0
#define pwm_channel_1  PORTBbits.RB1
#define pwm_channel_2  PORTBbits.RB2
#define pwm_channel_3  PORTBbits.RB3
#define pwm_channel_4  PORTBbits.RB4
#define pwm_channel_5  PORTBbits.RB5
#define pwm_channel_6  PORTBbits.RB6
#define pwm_channel_7  PORTBbits.RB7

struct pwm_output {
    unsigned char high;
    unsigned char low;
    unsigned char enabled;
} pwm_output_t;

unsigned char pwm_current;
struct pwm_output pwm_outputs[PWM_NUM_CHANNELS];

/* and macros to turn them on and off */
#define _on(a)	a = 1
#define _off(a)	a = 0

/***** Function Prototypes ***************************************************/
void highp_handler(void);
void lowp_handler(void);
void wait(void);

void pwm_changePosition(unsigned char channel, unsigned char value);
void handle_serial_rx(void);

/***** Setup for serial communications ***************************************/

#pragma idata bigdata

cbuffer_buff_t tx_buff, rx_buff;

char hexchar[] = "01234567890ABCDE";

#pragma code highp_interrupt = 0x08
void highp_int(void) {
    _asm goto highp_handler _endasm
}

#pragma code lowp_interrupt = 0x18
void lowp_int(void) {
    _asm goto lowp_handler _endasm
}

/***** Program Code **********************************************************/
#pragma code

/* High priority interrupt handler */
#pragma interrupt highp_handler
void highp_handler (void) {
    unsigned char c;

    /* Timer 0 interrupt: schedule next PWM event */
    if(INTCONbits.TMR0IF == 1) {
        /* Preload timer 0 so that it has a period of 3ms */
        TMR0L = 64;

        /* in case timer1 was delinquent */
        if( T1CONbits.TMR1ON == 1 ) {
        
            T1CONbits.TMR1ON = 0;   /* disable the timer */

            if(pwm_outputs[pwm_current].enabled == 1) {
                switch( pwm_current ) {
                    case 0: _off(pwm_channel_0); break;
                    case 1: _off(pwm_channel_1); break;
                    case 2: _off(pwm_channel_2); break;
                    case 3: _off(pwm_channel_3); break;
                    case 4: _off(pwm_channel_4); break;
                    case 5: _off(pwm_channel_5); break;
                    case 6: _off(pwm_channel_6); break;
                    case 7: _off(pwm_channel_7); break;
                    default: break;
                }
            }
        }

        pwm_current = (pwm_current + 1) % PWM_NUM_CHANNELS;
        if( pwm_outputs[pwm_current].enabled == 1) {
            /* Load timer1
             * The high byte is written first, but does not get updated until
             * the lower byte is written. */
            TMR1H = pwm_outputs[pwm_current].high;
            TMR1L = pwm_outputs[pwm_current].low;
            
            switch( pwm_current ) {
                case 0: _on(pwm_channel_0); break;
                case 1: _on(pwm_channel_1); break;
                case 2: _on(pwm_channel_2); break;
                case 3: _on(pwm_channel_3); break;
                case 4: _on(pwm_channel_4); break;
                case 5: _on(pwm_channel_5); break;
                case 6: _on(pwm_channel_6); break;
                case 7: _on(pwm_channel_7); break;
                default: break;
            }

            T1CONbits.TMR1ON = 1;   /* enable the timer */
        }

        /* clear timer0 interrupt flag */
        INTCONbits.TMR0IF = 0;
    }
    /* Timer 1 interrupt: turn off the current pin
     * and the timer1 interrupt 
     */
    else if( PIR1bits.TMR1IF == 1 ) {

        T1CONbits.TMR1ON = 0;   /* disable the timer */

        switch( pwm_current ) {
            case 0: _off(pwm_channel_0); break;
            case 1: _off(pwm_channel_1); break;
            case 2: _off(pwm_channel_2); break;
            case 3: _off(pwm_channel_3); break;
            case 4: _off(pwm_channel_4); break;
            case 5: _off(pwm_channel_5); break;
            case 6: _off(pwm_channel_6); break;
            case 7: _off(pwm_channel_7); break;
            default: break;
        }

        /* clear timer1 interrupt flag */
        PIR1bits.TMR1IF = 0;
    }

}

/* Low priority interrupt handler: handle serial communication */
#pragma interruptlow lowp_handler
void lowp_handler (void) {
    static char dat;

    if( PIR1bits.RCIF == 1) 
        /* just read the data and write it to the buffer... facile! */
        cbuffer_write_c( &rx_buff, ReadUSART() );

        /* should add a bit here to check if buffer is full for flow control */
    
    else if( PIR1bits.TXIF == 1 ) {
        if( cbuffer_read_c( &tx_buff, &dat ) == CBUFFER_MSG_SUCCESS ) 
            TXREG = dat;
        else 
            /* we cant simply shut off the TX here because the last character
             * is possibly still transferring.
             */
            if( TXSTAbits.TRMT == 1 )
                TXSTAbits.TXEN = 0;
    }
}

#define pwm_enable(channel)                         \
    if(channel >=0 && channel <PWM_NUM_CHANNELS)    \
        pwm_outputs[channel].enabled = 1;

#define pwm_disable(channel)                        \
    if(channel >=0 && channel <PWM_NUM_CHANNELS)    \
        pwm_outputs[channel].enabled = 0;

void pwm_changePosition(unsigned char channel, unsigned char value) {
    unsigned int newval;

    if(channel >= 0 && channel < PWM_NUM_CHANNELS) {
        newval = 65535 - 250 - (int)value*4;

        pwm_outputs[channel].high = (newval >> 8) & 255;
        pwm_outputs[channel].low = newval & 255;
    }
}

void handle_serial_rx(void) {
#define HANDLE_BUFFER_LENGTH 10
    static char buff[HANDLE_BUFFER_LENGTH];
    static unsigned char pos = 0;

    unsigned char i;
    unsigned char result;
    unsigned int newval;
    unsigned char outbuff[20];


    if( pos >= HANDLE_BUFFER_LENGTH )
        pos = 0;

    if( cbuffer_read_c( &rx_buff, &buff[pos] ) != CBUFFER_MSG_SUCCESS )
        return;
    pos++;

    /* Reset the buffer */
    if( buff[pos-1] == '+' )
        pos = 0;

    switch( buff[0] ) {
        case 'l':
            /* Control the LED
             * syntax: lx, x is 0 for off or 1 for on
             */
            if(pos > 1) {
                if(buff[1] - '0' == 0)
                    _off(LED);
                else
                    _on(LED);
                pos = 0;
            }
            break;
        case 'm':
            /* Change the servo output value
             * syntax: m#xxx, # is the channel number, xxx is the value in dec
             */
            if(pos > 4) {
                result = (buff[2] - '0') * 100 +
                         (buff[3] - '0') * 10 +
                         (buff[4] - '0');

                pwm_changePosition(buff[1] - '0', result);
                pos=0;
            }
            break;
        case 'e':
            /* Turn on (enable) the channel 
             * syntax: e#, # is the channel number
             */
            if(pos > 1) {
                pwm_enable(buff[1] - '0');
                pos = 0;
            }
            break;
        case 'd':
            /* turn off (disable) the channel
             * syntax: d#, # is the channel number
             */
            if(pos > 1) {
                pwm_disable(buff[1] - '0');
                pos=0;
            }
            break;

        default:
            /* we don't have a command in the first byte, so drop the buffer */
            pos=0;
            break;
    }
    return;
}

void main (void)
{
    unsigned char i;
    int wait_time;

    OSCCON  = 0x72;  /* Internal 8MHz osc */
	
/* set up serial port and high-priority interrupts for the serial port */
    baudUSART(BAUD_16_BIT_RATE &
              BAUD_WAKEUP_OFF &
              BAUD_AUTO_OFF);
			
    OpenUSART (USART_TX_INT_ON &
               USART_RX_INT_ON &
               USART_ASYNCH_MODE &
               USART_EIGHT_BIT &
               USART_CONT_RX &
               USART_BRGH_LOW, 51);

    IPR1bits.RCIP = 0;	/* Make receive interrupt low priority */
    IPR1bits.TXIP = 0;	/* Make transmit interrupt low priority */

/* clear circular buffer */
    cbuffer_init( &tx_buff );
    cbuffer_init( &rx_buff );


/* Set up timer0 to be a 4ms timer.  We really want it to fire at 3ms so that
 * 10 pwm channels can be handled each 30ms, so a preset value should be loaded
 * in the interrupt handler.
 */
    T0CONbits.T08BIT = 1;   /* Use 8-bit timer mode */
    T0CONbits.T0CS = 0;     /* Use the Fosc/4 as a clock source */
    T0CONbits.PSA = 0;      /* Use the prescaler */
    T0CONbits.T0PS2 = 1;    /* Set the prescaler value */
    T0CONbits.T0PS1 = 0;    /* 100 = /32 prescale */
    T0CONbits.T0PS0 = 0;

    INTCONbits.TMR0IE = 1;  /* turn on timer0 interrupt, this goes off when the
                             * timer rolls over 0xFFFF
                             */
    INTCON2bits.TMR0IP = 1; /* make timer0 high priority */


/* Set up timer1 to have a max period of 2ms */
    T1CONbits.RD16 = 0;     /* R/W in two 1 bit operations */
    T1CONbits.T1RUN = 0;    /* system clock status bit? */
    T1CONbits.T1CKPS1 = 1;  /* prescale select */
    T1CONbits.T1CKPS0 = 0;  /* 10 = /4 prescale */
    T1CONbits.T1OSCEN = 0;  /* disable the oscillator */
    T1CONbits.T1SYNC = 0;   /* ignored whe TMR1CS=0 */
    T1CONbits.TMR1CS = 0;   /* use fosc/4 */
    T1CONbits.TMR1ON = 1;   /* enable the timer */

    PIE1bits.TMR1IE = 1;    /* enable the timer1 interrupt, this goes off when
                             * the timer rolls over 0xFFFF
                             */
    IPR1bits.TMR1IP = 1;    /* make timer1 high priority */

/* Set up analog inputs */
#if 0
    PIE1bits.ADIE = 0;      /* turn off adc interrupt */
    PIR1bits.ADIF = 0;
    ADCON1 = 0b00001100;    /* AN0-AN3 are anaog inputs */
                            /* Voltage references use Vss, Vdd */
    ADCON0 = 0b00000001;    /* Select AN0, but don't do anything yet */
    ADCON2 = 0b00011101;    /* Left Justified, 6 TAD, Fosc/16 */
#endif

/* Set up port A */
    ADCON1 = 0b00001111;    /* turn off ADC functionality */
    TRISA  = 0b00000000;    /* 0=output, 1=input */
    PORTA  = 0b00000000;    /* value of the output pins */

/* Set up port B */
    TRISB = 0b00000000;         /* 0=output, 1=input */
    PORTB = 0b00000000;         /* value of the output pins */
    INTCON2bits.NOT_RBPU = 0;   /* Turn on pull-up resistors for inputs */

#if 0
    INTCON2bits.RBIP = 1;  /* Make port B interrupt priority high */
    INTCONbits.RBIE = 1;   /* Enable port B interrupts */
#endif
    
    /* init our pwm driver values
     * neutral corresponds to 2^16-625-128 counts */ 
    for(i=0; i<10; i++) {
        pwm_outputs[i].high = 253;
        pwm_outputs[i].low = 15;
        pwm_outputs[i].enabled = 0;
    }

/* Turn on interrupts  */
    RCONbits.IPEN = 1;    /* Enable interrupt priority */
    INTCONbits.GIEH = 1;  /* Enable all high priority interrupts */
    INTCONbits.GIEL = 1;  /* Enable all low priority interrupts */

/* turn on the status LED to signal that the system is up */
    _on(LED);

/* Main loop */
    while (1)
    {
        for(wait_time=0;wait_time<200;wait_time++) {}
           
        if( cbuffer_hasData( &rx_buff ) == CBUFFER_MSG_HAS_DATA )
            handle_serial_rx();
    }
}
